require "util"
--require "defines"

local PollutionWeight   = 7.0
local SpawnerWeight     = 6.0
local TimeWeight        = 2.0
local TotalWeight       = 10.0 -- 15/10, so some overlap intentional

local PollutionCap  = 250 * 1000 * 1000 -- 250 Million. Frontloaded by default, so it still grows really fast early on.

local TimeCap = 24 * 60 * 60 * 60 -- After 24 hours, hits cap, and stops increasing. Irreversable.

local SpawnerCap = 120         -- 120 Spawner deaths for maximum evolution factor.
local SpawnerDecay = 1/(60*12) -- 1 per 12 minutes, 1/3600, 5/Hour, 24 hours required for full decay.

local pollutionExponent = 0.5 

local spawnerExponent = 0.4   

local timeExponent = 0.4    
  
-- For exponents. Any factor works, the further from 1.0, the larger the difference. At 0% and 100%, is is always 0% and 100%.
-- 0.5 is 2x Frontloaded. Grows fast, then slows down.                           1% ->   10%, 25% -> 50%. 50% -> 70%, 90% -> 94%
-- 1.0 is Linear. Constant growth the whole way.                                 1% ->    1%, 50% -> 50%, 90% -> 90%
-- 2.0 is 2x Rearloaded. Grows very slowly, then goes fast at the end.           1% -> 0.01%, 10% ->  1%, 50% -> 25% 90% -> 81%
-- Vanilla is essentially frontloaded, but done somewhat differently.  Never can reach 100%. 
-- 100% -> 50%, 200% -> 66%, 300% -> 75%, 1,000% -> 90%. 10,000% -> 99%, 100,000% -> 99.9%, 1,000,000% -> 99.99%


-- Need Power Values

script.on_init(function()
   global.pollution_spawner = 0
   game.map_settings.enemy_evolution.time_factor = 0.000008 * 0
   game.map_settings.enemy_evolution.pollution_factor = 0.00003 * 0
end)

script.on_configuration_changed(function(data)
	if data.mod_changes ~= nil and data.mod_changes["Rescaled-Evolution-Factor"] ~= nil and data.mod_changes["Rescaled-Evolution-Factor"].old_version == nil then
		global.pollution_spawner = 0
		game.map_settings.enemy_evolution.time_factor = 0.000008 * 0
		game.map_settings.enemy_evolution.pollution_factor = 0.00003 * 0
	end
 end)
 
 script.on_load(function()
	if remote.interfaces.EvoGUI ~= nil then registriere_sensoren() end
 end)

function pollutiontest(event)
    if game.tick % 60 == 24 then
		if remote.interfaces.EvoGUI ~= nil then registriere_sensoren() end

		local summe = 0 -- Pollution Count
		local anzahl = 0 -- ChunkCount
      
		local s = game.surfaces["nauvis"]
	  
		for c in s.get_chunks() do                             -- Calculate how much pollution is on the game's primary surface.
			summe = summe + s.get_pollution({c.x *32,c.y *32}) -- Might work wrongly if a mod messes with many surfaces/locks pollution underground or something, 
			anzahl = anzahl + 1                                -- But that is outside the scope of this mod.
		end
	  
		if anzahl < 1000 then --Mindestwert für die anzahl, erleichtert den Start.
			anzahl = 1000     --Minimum value for the number , facilitates the start
		end                   --Until this many chunks are loaded, assume you have at least 1000. Past that, reduces threshhold.
	                          -- Currently disabled. I can't see any reason to have your exploration reduce how much biters hate pollution.
        
		--local pollution_summe = math.pow((summe/math.log(anzahl))/PollutionCap,pollutionExponent) * PollutionWeight --(Weight = 50%)
        
		local pollution_summe = math.pow(summe/PollutionCap,pollutionExponent) * PollutionWeight
	  
		if pollution_summe > PollutionWeight then -- Factor Directly from Pollution.
			pollution_summe = PollutionWeight     -- Any increases or decreases in Global pollution cause instant shifts in evolution factor.
		end                                       -- Hence, going green, or shutting down your factory can drop the Evo factor significantly.
	  
		local vergangene_zeit = TimeWeight
		if game.tick < TimeCap then -- Factor from time elapsed. Cannot be reversed.
			vergangene_zeit = math.pow(game.tick / TimeCap, timeExponent) * TimeWeight
		end
        
		local pollution_spawner_anzahl = SpawnerWeight
		if global.pollution_spawner < SpawnerCap then -- Factor from Spawners killed. Can decay slowly, but tends to linger for an excessively long time.
			pollution_spawner_anzahl = math.pow(global.pollution_spawner / SpawnerCap,spawnerExponent) * SpawnerWeight
		end
        
	  	if pollution_spawner_anzahl > SpawnerWeight then -- Make sure it never goes above the cap.
			pollution_spawner_anzahl = SpawnerWeight
		end
	  
		local factor = (pollution_summe + pollution_spawner_anzahl + vergangene_zeit) / TotalWeight --Again "entwichten". " to escape, to get away " I don't understand.
        
		if factor > 1 then -- Can't pass 1.0, or 100% evolution factor, no matter how much you piss them off.
			factor = 1
		end
	  
		game.evolution_factor = factor -- Entirely override the game's normal factor. This means that any other mods trying to manipulate it will fail miserably.
 		
		if remote.interfaces.EvoGUI ~= nil then -- Send data to EoGUI, if installed. Hard to see what the mod does without it, honestly.
			remote.call("EvoGUI", "update_remote_sensor", "rescaled_evolution_factor_sensor_pollution", {"", {"rescaled_evolution_factor_sensor_pollution_prefix"}, " ", summe_format(summe)})
			remote.call("EvoGUI", "update_remote_sensor", "rescaled_evolution_factor_sensor_spawner", {"", {"rescaled_evolution_factor_sensor_spawner_prefix"}, " ", math.ceil(global.pollution_spawner)})
			remote.call("EvoGUI", "update_remote_sensor", "rescaled_evolution_factor_sensor_anteile", {"", {"rescaled_evolution_factor_sensor_anteile_prefix"}, " ", string.format("%.1f", vergangene_zeit * 10).."%|", string.format("%.1f", pollution_spawner_anzahl * 10).."%|", string.format("%.1f", pollution_summe * 10).."%"})
		end
	end
   
    if (game.tick % 60 == 25 and global.pollution_spawner > 0) then -- Slowly reduce anger from spawner death, constantly. Obviously it can't go negative.
		global.pollution_spawner = math.max(global.pollution_spawner - SpawnerDecay,0)   --Vergiss mit der Zeit alte Fehden.
                                                                    --Forget the time old feuds .
	end
end

function spawner_death(event)
    if (global.pollution_spawner < SpawnerCap) and (event.entity.type == "unit-spawner") then
        global.pollution_spawner = global.pollution_spawner + 1 -- Each spawner death makes them angrier.
    else
        -- In the future, check for Worms, add something like 10% worth in that case.
    end   
end

function registriere_sensoren()
	if global.ist_registriert == nil then -- Uhm. I think this handles registering EvoGUI stuff.
		local sensor_data = {mod_name = "Evolve Factor",
							 name = "rescaled_evolution_factor_sensor_pollution",
							 text = {"rescaled_evolution_factor_sensor_pollution_display", 42},
							 caption = {"rescaled_evolution_factor_sensor_pollution_caption"}}
   
		remote.call("EvoGUI", "create_remote_sensor", sensor_data)
   
		sensor_data = {mod_name = "Evolve Factor",
					   name = "rescaled_evolution_factor_sensor_spawner",
					   text = {"rescaled_evolution_factor_sensor_spawner_display", 42},
					  caption = {"rescaled_evolution_factor_sensor_spawner_caption"}}
   
		remote.call("EvoGUI", "create_remote_sensor", sensor_data)
   
		sensor_data = {mod_name = "Evolve Factor",
					   name = "rescaled_evolution_factor_sensor_anteile",
					   text = {"rescaled_evolution_factor_sensor_anteile_display", 42},
					   caption = {"rescaled_evolution_factor_sensor_anteile_caption"}}
   
		remote.call("EvoGUI", "create_remote_sensor", sensor_data)
		
		global.ist_registriert = true
	end
end

function summe_format(v) -- Some kind of formatting thing?
	local s = string.format("%d", math.floor(v))
	local pos = string.len(s) % 3
	if pos == 0 then pos = 3 end
	return string.sub(s, 1, pos) .. string.gsub(string.sub(s, pos+1), "(...)", " %1")
end

function prozent_format(v) -- Isn't even used?
    local whole_number = math.floor(v)
    local fractional_component = math.floor((v - whole_number) * 10)
	
	return string.format("%d.%d%%", whole_number, fractional_component)
end

script.on_event(defines.events.on_tick, pollutiontest)
script.on_event(defines.events.on_entity_died , spawner_death)